/* AbiSource Program Utilities
 * Copyright (C) 1998 AbiSource, Inc.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 */
 


#include <stdio.h>
#include <string.h>

#include "ut_types.h"
#include "ut_assert.h"
#include "ut_debugmsg.h"
#include "ut_string.h"
#include "ps_Font.h"
#include "ps_Generate.h"
#include "ps_Graphics.h"
#include "ps_Fontmap.h"

// the resolution that we report to the application (pixels per inch).
#define PS_RESOLUTION		7200


/*****************************************************************
******************************************************************
** This file contains the Interface between the application and
** the PostScript generation code.
******************************************************************
*****************************************************************/

PS_Graphics::PS_Graphics(const char * szFilename,
						 const char * szTitle,
						 const char * szSoftwareNameAndVersion,
						 UT_Bool	  bIsFile)
{
	UT_ASSERT(szFilename && *szFilename);
	m_szFilename = szFilename;
	m_szTitle = szTitle;
	m_szSoftwareNameAndVersion = szSoftwareNameAndVersion;
	m_pCurrentFont = 0;
	m_bStartPrint = UT_FALSE;
	m_bStartPage = UT_FALSE;
	m_bNeedStroked = UT_FALSE;
	m_bIsFile = bIsFile;
	m_ps = 0;
	m_fm = 0;
}

PS_Graphics::~PS_Graphics()
{
	// TODO free stuff
}

UT_Bool PS_Graphics::queryProperties(DG_Graphics::Properties gp) const
{
	switch (gp)
	{
	case DGP_SCREEN:
		return UT_FALSE;
	case DGP_PAPER:
		return UT_TRUE;
	default:
		UT_ASSERT(0);
		return UT_FALSE;
	}
}

void PS_Graphics::setFont(DG_Font* pFont)
{
	UT_ASSERT(pFont);
	ps_Font * pNewFont = (static_cast<ps_Font*> (pFont));

	// TODO Not always what we want, i.e., start of a new page.
	// TODO I added a call directly to _startPage to call _emit_SetFont();
	// TODO I would rather do it all here.
	if (pNewFont == m_pCurrentFont ) 
		return;
	
	m_pCurrentFont = pNewFont;
	if (m_ps)
		_emit_SetFont();
}

UT_uint32 PS_Graphics::getFontAscent()
{
	UT_ASSERT(m_pCurrentFont);

	GlobalFontInfo * gfi = m_pCurrentFont->m_afm_fi->gfi;
	UT_ASSERT(gfi);

	return _scale(gfi->ascender);
}

UT_uint32 PS_Graphics::getFontDescent()
{
	UT_ASSERT(m_pCurrentFont);

	GlobalFontInfo * gfi = m_pCurrentFont->m_afm_fi->gfi;
	UT_ASSERT(gfi);

	UT_ASSERT(gfi->descender <= 0);
	return _scale(-gfi->descender);
}

UT_uint32 PS_Graphics::getFontHeight()
{
	UT_ASSERT(m_pCurrentFont);

	GlobalFontInfo * gfi = m_pCurrentFont->m_afm_fi->gfi;
	UT_ASSERT(gfi);

	// TODO who takes care of leading ??
	
	UT_ASSERT(gfi->descender <= 0);
	return _scale(gfi->ascender - gfi->descender);
}
	
UT_uint32 PS_Graphics::measureString(const UT_UCSChar* s, int iOffset, 
									int num, unsigned short* pWidths)
{
	UT_ASSERT(m_pCurrentFont);
	UT_ASSERT(s);
	const UT_UCSChar * p = s+iOffset;
	const int * cwi = m_pCurrentFont->m_afm_fi->cwi;

	UT_uint32 iCharWidth = 0;
	for (int k=0; k<num; k++)
	{
		UT_ASSERT(p[k] < 256);			// TODO deal with Unicode
		register int x = _scale(cwi[p[k]]);
		
		iCharWidth += x;
		pWidths[k] = x;
	}
		
	return iCharWidth;
}
	
UT_uint32 PS_Graphics::getResolution() const
{
	return PS_RESOLUTION;
}

void PS_Graphics::setColor(UT_RGBColor& /*clr*/)
{
}

DG_Font* PS_Graphics::findFont(const char* pszFontFamily, 
							   const char* pszFontStyle, 
							   const char* pszFontVariant, 
							   const char* pszFontWeight, 
							   const char* /*pszFontStretch*/, 
							   const char* pszFontSize)
{
	UT_sint32 iSize = convertDimension(pszFontSize); // in pixels at our resolution

	UT_uint32 k;
	UT_uint32 kLimit = m_vecFontList.getItemCount();
	
	for (k=0; k<kLimit; k++)
	{
		ps_Font * psf = (ps_Font *)m_vecFontList.getNthItem(k);
		UT_ASSERT(psf);
		if (psf->matchesFont(pszFontFamily,pszFontVariant,pszFontStyle,pszFontWeight, iSize))
			return psf;
	}
	// font not found in list, create a new one.
	UT_uint32 ndx = m_vecFontList.getItemCount();
	ps_Font * ps = new ps_Font(ndx);
	if (!ps)
		return 0;

	char *bufFilename = new char[1024];
	if(!m_fm)
	{
		m_fm = new ps_Fontmap();
		m_fm->createFontHash();
		if(!m_fm)
			return UT_FALSE;
	}

	// TODO remember all the args to this function
	// TODO verify that variant maps to face
	m_fm->mapFontName(pszFontFamily, pszFontVariant, pszFontStyle, pszFontWeight, bufFilename, 1024 );
	ps->setFont(pszFontFamily,pszFontVariant,pszFontStyle,pszFontWeight,iSize);
	if (!ps->loadFont(bufFilename))
	{
		delete ps;
		return 0;
	}

	if(bufFilename)
		delete bufFilename;
	
	m_vecFontList.addItem(ps);
	return static_cast<DG_Font*> (ps);
}

#define OUR_LINE_LIMIT		70			/* DSC3.0 limits to 256 */

void PS_Graphics::drawChars(const UT_UCSChar* pChars, int iCharOffset, 
							int iLength, UT_sint32 xoff, UT_sint32 yoff)
{
	UT_ASSERT(m_pCurrentFont);
	
	char buf[OUR_LINE_LIMIT*2];
	const UT_UCSChar * pS = pChars+iCharOffset;
	const UT_UCSChar * pEnd = pS+iLength;
	char * pD = buf;

	*pD++ = '(';
	while (pS<pEnd)
	{
		if (pD-buf > OUR_LINE_LIMIT)
		{
			*pD++ = '\\';
			*pD++ = '\n';
			*pD++ = 0;
			m_ps->writeBytes(buf);
			pD = buf;
		}

		// TODO deal with Unicode issues.

		switch (*pS)
		{
		case 0x08:		*pD++ = '\\';	*pD++ = 'b';	break;
		case 0x09:		*pD++ = '\\';	*pD++ = 't';	break;
		case 0x0a:		*pD++ = '\\';	*pD++ = 'n';	break;
		case 0x0c:		*pD++ = '\\';	*pD++ = 'f';	break;
		case 0x0d:		*pD++ = '\\';	*pD++ = 'r';	break;
		case '\\':		*pD++ = '\\';	*pD++ = '\\';	break;
		case '(':		*pD++ = '\\';	*pD++ = '(';	break;
		case ')':		*pD++ = '\\';	*pD++ = ')';	break;
		default:		*pD++ = (char)*pS;				break;
		}
		pS++;
	}
	*pD++ = ')';
	sprintf(pD," %ld %ld MS\n",xoff,yoff);
	m_ps->writeBytes(buf);
}

void PS_Graphics::drawLine(UT_sint32 x1, UT_sint32 y1, UT_sint32 x2, UT_sint32 y2)
{
	// TODO This is used for lines in the document, as well as underlines
	// TODO and strikes.
	m_bNeedStroked = UT_TRUE;

	char buf[OUR_LINE_LIMIT*2];
	UT_sint32 nA = getFontAscent();
	sprintf(buf,"%ld %ld %ld %ld ML\n", x2, y2-nA, x1, y1-nA);
	m_ps->writeBytes(buf);
}

void PS_Graphics::xorLine(UT_sint32, UT_sint32, UT_sint32, UT_sint32)
{
}

void PS_Graphics::fillRect(UT_RGBColor& /*c*/, UT_sint32 /*x*/, UT_sint32 /*y*/, UT_sint32 /*w*/, UT_sint32 /*h*/)
{
}

void PS_Graphics::invertRect(const UT_Rect* /*pRect*/)
{
	// invertRect is only used for drawing the selection region, which does not apply on paper.
	UT_ASSERT(UT_SHOULD_NOT_HAPPEN);
}

void PS_Graphics::setClipRect(const UT_Rect* /*pRect*/)
{
	// setClipRect is only used for handling screen exposes, which do not happen on paper.
	UT_ASSERT(UT_SHOULD_NOT_HAPPEN);
}

void PS_Graphics::clearArea(UT_sint32 /*x*/, UT_sint32 /*y*/,
							UT_sint32 /*width*/, UT_sint32 /*height*/)
{
	// clearArea is only used for re-drawing edited regions, which does not apply on paper.
	UT_ASSERT(UT_SHOULD_NOT_HAPPEN);

}

void PS_Graphics::scroll(UT_sint32, UT_sint32)
{
	UT_ASSERT(UT_SHOULD_NOT_HAPPEN);
}

UT_Bool PS_Graphics::startPrint(void)
{
	UT_ASSERT(!m_bStartPrint);
	m_bStartPrint = UT_TRUE;
	m_ps = new ps_Generate(m_szFilename);
	if (!m_ps)
		return UT_FALSE;

	return _startDocument();
}

UT_Bool PS_Graphics::startPage(const char * szPageLabel, UT_uint32 pageNumber,
							   UT_Bool bPortrait, UT_uint32 iWidth, UT_uint32 iHeight)
{
	if (m_bStartPage)
		_endPage();
	m_bStartPage = UT_TRUE;
	m_bNeedStroked = UT_FALSE;
	return _startPage(szPageLabel,pageNumber,bPortrait,iWidth,iHeight);
}

UT_Bool PS_Graphics::endPrint(void)
{
	if (m_bStartPage)
		_endPage();
	return _endDocument();
}

/*****************************************************************/
/*****************************************************************/

UT_uint32 PS_Graphics::_scale(UT_uint32 units) const
{
	// we are given a value from the AFM file which are
	// expressed in 1/1000ths of the scaled font.
	// return the number of pixels at our resolution.

	return (units * m_pCurrentFont->m_iSize / 1000);
}

UT_Bool PS_Graphics::_startDocument(void)
{
	// open the file and begin writing the prolog
	// return false if an error occurs.

	//////////////////////////////////////////////////////////////////
	// DSC3.0/Prolog/Header
	//////////////////////////////////////////////////////////////////
	
	if (!m_ps->openFile(m_bIsFile))
		return UT_FALSE;

	// we use the argv-version of these so that we get PS-escaping on
	// the strings (which will probably have spaces in them).
	
	if (m_szSoftwareNameAndVersion && *m_szSoftwareNameAndVersion)
		m_ps->formatComment("Creator",&m_szSoftwareNameAndVersion,1);
	if (m_szTitle && *m_szTitle)
		m_ps->formatComment("Title",&m_szTitle,1);

	_emit_DocumentNeededResources();
	
	// TODO add other header-comments here
	
	m_ps->formatComment("EndComments");

	///////////////////////////////////////////////////////////////////
	// DSC3.0/Prolog/ProcedureDefinitions
	///////////////////////////////////////////////////////////////////
		
	m_ps->formatComment("BeginProlog");

	_emit_PrologMacros();
	_emit_FontMacros();
	
	// TODO add rest of prolog

	m_ps->formatComment("EndProlog");

	///////////////////////////////////////////////////////////////////
	// DSC3.0/Script/DocumentSetup
	///////////////////////////////////////////////////////////////////

	m_ps->formatComment("BeginSetup");

	_emit_IncludeResource();
	
	// TODO add other setup stuff

	m_ps->formatComment("EndSetup");
	return UT_TRUE;
}

UT_Bool PS_Graphics::_startPage(const char * szPageLabel, UT_uint32 pageNumber,
								UT_Bool bPortrait, UT_uint32 iWidth, UT_uint32 iHeight)
{
	// emit stuff prior to each page
	
	char buf[1024];
	sprintf(buf,"%ld",pageNumber);

	const char * argv[2] = { szPageLabel, buf };
	m_ps->formatComment("Page",argv,2);

	// TODO add other page-header comments here

	m_ps->formatComment("BeginPageSetup");

	sprintf(buf,"%ld %ld %d %s\n",iWidth,iHeight,PS_RESOLUTION,((bPortrait) ? "BPP" : "BPL"));
	m_ps->writeBytes(buf);
	
	// TODO add page-setup stuff here

	m_ps->formatComment("EndPageSetup");

	// Note, the actual PS for the page will be sent by the
	// individual drawing routings following the EndPageSetup.

	// Note, need to reset font at the beginning of each page
	_emit_SetFont();

	return UT_TRUE;
}

UT_Bool PS_Graphics::_endPage(void)
{
	// emit stuff following each page

	m_ps->formatComment("PageTrailer");

	if(m_bNeedStroked)
		m_ps->writeBytes("stroke\n");
	m_ps->writeBytes("EP\n");

	// TODO add any page-trailer stuff here
	// TODO (this inludes an atend's that we used in the page header)

	// Note, either the next page or the document-trailer will follow this.
	return UT_TRUE;
}

UT_Bool PS_Graphics::_endDocument(void)
{
	// emit the document trailer

	m_ps->formatComment("Trailer");
	
	// TODO add any trailer stuff here
	// TODO (this includes an atend's that we used in the document header)

	UT_Bool bStatus = m_ps->formatComment("EOF");

	m_ps->closeFile();
	delete m_ps;
	m_ps = 0;

	return bStatus;
}

void PS_Graphics::_emit_DocumentNeededResources(void)
{
	UT_Vector vec;
	UT_uint32 k;
	UT_uint32 kLimit = m_vecFontList.getItemCount();

	for (k=0; k<kLimit; k++)
	{
		vec.addItem("font");
		ps_Font * psf = (ps_Font *)m_vecFontList.getNthItem(k);
		vec.addItem(psf->m_afm_fi->gfi->fontName);
	}

	// TODO add any other resources here
	
	m_ps->formatComment("DocumentNeededResources",&vec);
}

void PS_Graphics::_emit_IncludeResource(void)
{
	UT_uint32 k;
	UT_uint32 kLimit = m_vecFontList.getItemCount();

	for (k=0; k<kLimit; k++)
	{
		ps_Font * psf = (ps_Font *)m_vecFontList.getNthItem(k);
		m_ps->formatComment("IncludeResource",psf->m_afm_fi->gfi->fontName);
	}

	// TODO add any other IncludeResource's here
}

#define NrElements(a)	(sizeof(a)/sizeof(a[0]))

void PS_Graphics::_emit_PrologMacros(void)
{
	static const char * t[] = {
		"/FSF {findfont exch scalefont} bind def",			// Find & scale a font. <size> /<fontname> FSD
		"/MS  {neg moveto show} bind def",					// Move and draw. (<string>) <x> <y> MS
		"/BP  {gsave 72 exch div dup scale} bind def",		// Begin Page. <res> BP
		"/SZ  {/HH exch def /WW exch def} bind def",		// Page Size.  <w> <h> SZ
		"/BPP {BP SZ 0 HH translate} bind def",				// Begin Portrait Page.  <w> <h> <res> BPP
		"/BPL {BP SZ 90 rotate} bind def",					// Begin Landscape Page. <w> <h> <res> BPP
		"/EP  {grestore showpage} bind def",				// EP
		"/ML  {neg moveto neg lineto} bind def",			// Move and Line. <x2> <y2> <x1> <y1> ML
	};

	char buf[1024];
	
	for (unsigned int k=0; k<NrElements(t); k++)
	{
		sprintf(buf,"  %s\n",t[k]);
		m_ps->writeBytes(buf);
	}
}

void PS_Graphics::_emit_FontMacros(void)
{
	char buf[1024];
	UT_uint32 k;
	UT_uint32 kLimit = m_vecFontList.getItemCount();

	for (k=0; k<kLimit; k++)
	{
		ps_Font * psf = (ps_Font *)m_vecFontList.getNthItem(k);
		sprintf(buf,"  /F%ld {%ld /%s FSF setfont} bind def\n",k,psf->m_iSize,psf->m_afm_fi->gfi->fontName);
		m_ps->writeBytes(buf);
	}
}

void PS_Graphics::_emit_SetFont(void)
{
	char buf[1024];
	sprintf(buf,"F%ld\n",m_pCurrentFont->m_ndx);
	m_ps->writeBytes(buf);
}
